proc exiterror {msg} {
	return $msg $::errorInfo
}

proc chr_clip chr {
	regsub {^[Cc][Hh][Rr]-?} $chr {} chr
	if {$chr eq "MT"} {set chr M}
	return $chr
}

proc tsv_basicfields {header {num 6} {giveerror 1}} {
	set poss [list_cor $header {chromosome begin end type ref alt}]
	set nfposs [list_find $poss -1]
	foreach nfpos $nfposs {
		switch $nfpos {
			0 {
				foreach name {chrom chr chr1 genoName tName contig} {
					set v [lsearch $header $name]
					if {$v != -1} break
				}
			}
			1 {
				foreach name {start end1 chromStart genoStart tStart txStart} {
					set v [lsearch $header $name]
					if {$v != -1} break
				}
			}
			2 {
				foreach name {start2 chromEnd genoEnd tEnd txEnd} {
					set v [lsearch $header $name]
					if {$v != -1} break
				}
			}
			4 {
				set v [lsearch $header reference]
			}
			5 {
				set v [lsearch $header alternative]
			}
			default {
				continue
			}
		}
		lset poss $nfpos $v
	}
	incr num -1
	set poss [lrange $poss 0 $num]
	set pos [lsearch $poss -1]
	if {$pos == 1} {
		set pos [lsearch $header pos]
		if {$poss == -1} {set pos [lsearch $header offset]}
		lset poss 1 $pos
		lset poss 2 $pos
		set pos [lsearch $poss -1]
	}
	if {$giveerror && ($pos != -1)} {
		set notfound [list_sub {chromosome begin end type ref alt} [list_find $poss -1]]
		error "header error: fields (or alternatives) not found: $notfound"
	}
	return $poss
}

proc gzopen {file {pos -1}} {
	if {![file exists $file]} {
		exiterror "Error: couldn't open \"$file\": no such file or directory"
	}
	if {[inlist {.rz} [file extension $file]]} {
		if {$pos == -1} {
			set f [open "| razip -d -c $file"]
		} else {
			set f [open "| razip -d -c -b $pos $file"]
		}
	} elseif {[inlist {.lz4} [file extension $file]]} {
		if {$pos == -1} {
			set f [open "| lz4c -d -c $file"]
		} else {
			error "positioning not supported in lz4 files"
		}
	} elseif {[inlist {.bgz .gz} [file extension $file]]} {
		if {$pos == -1} {
			set f [open "| zcat $file"]
		} else {
			error "positioning not supported in (b)gz files"
		}
	} elseif {[inlist {.bz2} [file extension $file]]} {
		if {$pos == -1} {
			set f [open "| bzcat $file"]
		} else {
			error "positioning not supported in bz2 files"
		}
	} else {
		set f [open $file]
		if {$pos != -1} {
			seek $f $pos
		}
	}
	return $f
}

proc gzclose {f} {
	if {[catch {close $f} error]} {
		if {$error eq "child killed: write on pipe with no readers"} return
		if {[regexp {Successfully decoded [0-9]+ bytes} $error]} return
		error $error
	}
}

proc tsv_open {f {keepheaderVar {}}} {
	if {$keepheaderVar ne ""} {
		upvar $keepheaderVar keepheader
	}
	set keepheader {}
	set keep 0
	set buffering [fconfigure $f -buffering]
	fconfigure $f -buffering line
	set split 1
	set line [gets $f]
	if {[regexp {^@HD[\t]VN} $line]} {
		# sam file
		while {![eof $f]} {
			set fchar [string index $line 0]
			if {$fchar ne "@"} {
				break
			}
			lappend keepheader \#$line
			set pos [tell $f]
			set line [gets $f]
		}
		set keepheader [join $keepheader \n]\n
		seek $f $pos start
		set len [llength [split $line \t]]
		set header {qname flag rname pos mapq cigar rnext pnext tlen seq qual}
		set size [expr {$len-11}]
		for {set i 1} {$i <= $size} {incr i} {
			lappend header opt$i
		}
		return $header
	}
	set fchar [string index $line 0]
	set fchar2 [string index $line 1]
	if {[regexp {##fileformat=VCF} $line]} {
		set vcf 1
	} else {
		set vcf 0
	}
	while {![eof $f]} {
		if {![string length $line]} {
			lappend keepheader \#
			set line [gets $f]
			break
		}
		set fchar [string index $line 0]
		if {$fchar eq ">"} {
			break
		} elseif {$fchar ne "#"} {
			break
		} elseif {$vcf} {
			set fchar2 [string index $line 1]
			if {$fchar2 ne "#"} {
				set split 0
				break
			}
		}
		lappend keepheader $line
		set line [gets $f]
	}
	fconfigure $f -buffering $buffering
	set fchar [string index $line 0]
	if {[inlist {# >} $fchar]} {
		set keepheader [join $keepheader \n]\n
		if {$vcf} {append keepheader \#}
		if {!$split} {
			return [string range $line 1 end]
		} else {
			return [split [string range $line 1 end] \t]
		}
	} else {
		if {[llength $keepheader]} {
			set keepheader [join $keepheader \n]\n
		}
		return [split $line \t]
	}
}

proc cg_fas2ifas {srcfile destfile} {
	# make new one
	set max [file size $srcfile]
	set f [gzopen $srcfile]
	set o [open $destfile.temp w]
	set line [gets $f]
	if {[string index $line 0] ne ">"} {
		error "$srcfile is not a FASTA file"
	}
	set ids {}
	puts $o $line
	set curid [string range $line 1 end]
	log "Converting genome: checking chromosome $curid"
	lappend ids $curid
	set a(s,$curid) [tell $o]
	while {![eof $f]} {
		set line [gets $f]
		if {$line eq ""} continue
		if {[string index $line 0] eq ">"} {
			set a(e,$curid) [tell $o]
			puts $o \n$line
			set curid [string range $line 1 end]
			log "Converting genome: checking chromosome $curid"
			lappend ids $curid
			set a(s,$curid) [tell $o]
		} else {
			puts -nonewline $o $line
		}
	}
	puts $o ""
	log "chromosome $curid finished"
	set a(e,$curid) [tell $o]
	close $o
	close $f
	# sort
	set sids [ssort -natural $ids]
	if {$sids ne $ids} {
		set f [open $destfile.temp]
		set o [open $destfile.temp2 w]
		set oi [open $destfile.index.temp2 w]
		foreach name $sids {
			log "Converting genome: writing chromosome $name"
			if {![regexp {chromosome ([^ ,]+)[ ,]} $name temp chr]} {
				if {![regexp {chr([^ ,]+)} $name temp chr]} {
					set chr [lindex $name end]
				}
			}
			set len [expr {$a(e,$name) - $a(s,$name)}]
			puts $o "\>$name"
			puts $oi "$chr\t[tell $o] $len"
			seek $f $a(s,$name)
			fcopy $f $o -size [expr {$len+1}]
		}
 		close $oi
		close $o
		close $f
		file rename -force $destfile.temp2 $destfile
		file rename -force $destfile.index.temp2 $destfile.index
		file delete $destfile.temp
	} else {
		set oi [open $destfile.index.temp w]
		foreach name $sids {
			if {![regexp {chromosome ([^ ,]+)[ ,]} $name temp chr]} {
				if {![regexp {chr([^ ,]+)} $name temp chr]} {
					set chr [lindex $name end]
				}
			}
			set len [expr {$a(e,$name) - $a(s,$name)}]
			puts $oi "$chr\t$a(s,$name) $len"
		}
 		close $oi
		file rename -force $destfile.temp $destfile
		file rename -force $destfile.index.temp $destfile.index
	}
}

proc genome_makefastaindex {fastafile} {
	log "Making index for $fastafile"
	if {[file isdir $fastafile]} {
		error "error: genome file given ($fastafile) is a directory"
	}
	set f [open $fastafile]
	set max [file size $fastafile]
	set keepmax [progress_max]
	progress_max $max 0
	set oi [open $fastafile.index.temp w]
	while {![eof $f]} {
		set line [gets $f]
		if {$line eq ""} continue
		if {[string index $line 0] ne ">"} {
			close $f ; close $oi
			set answer [::tk_messageBox -type okcancel -icon warning -message "The genome file given ($fastafile) is not in the correct format, will try to convert (old file is renamed to $fastafile.old)" -title Warning]
			if {$answer eq "cancel"} return
			log "Converting $fastafile to ifas, old rename to $fastafile.old"
			cg_fas2ifas $fastafile $fastafile.temp
			file rename -force $fastafile $fastafile.old
			file rename -force $fastafile.temp $fastafile
			log "Conversion $fastafile to ifas done, old renamed to $fastafile.old"
			error "Conversion $fastafile to ifas done, old renamed to $fastafile.old. Restart analysis"
		}
		set name [string range $line 1 end]
		if {![regexp {chromosome ([^ ,]+)[ ,]} $name temp chr]} {
			if {![regexp {chr([^ ,]+)} $name temp chr]} {
				set chr [lindex $name end]
			}
		}
		set start [tell $f]
		set seq [gets $f]
		set seqlen [string length $seq]
		progress_set $start
		log "Indexing $fastafile: chromosome $name"
		puts $oi "$chr\t$start $seqlen"
	}
	close $oi
	close $f
	file rename -force $fastafile.index.temp $fastafile.index
	progress_max $keepmax 0
}

proc genome_open {file} {
	global genomefasta
	if {[file isdir $file]} {
		set file [lindex [glob $file/genome_*.ifas] 0]
	}
	if {![file exists $file.index]} {
		genome_makefastaindex $file
	}
	set f [open $file]
	set fastaindex {}
	foreach {name data} [split [string trim [file_read $file.index]] \t\n] {
		if {![regexp {chr([^ ,]+)} $name temp chr]} {set chr [lindex $name end]}
		dict set fastaindex $chr $data
	}
	set genomefasta($f) $fastaindex
	return $f
}

proc genome_close {f} {
	global genomefasta
	close $f
	unset genomefasta($f)
}

proc genome_get {f chr start end {correctend 0}} {
	global genomefasta
	if {$end < $start} {error "end ($end) is smaller than start ($start)"}
	set fastaindex $genomefasta($f)
	if {[catch {
		set temp [dict get $fastaindex $chr]
	}]} {
		set chr [chr_clip $chr]
		if {[catch {set temp [dict get $fastaindex $chr]}]} {
			error "chromosome $chr not found in genome file"
		}
	}
	foreach {gstart glen} $temp break
	if {$end > $glen} {
		if {!$correctend} {
			error "trying to get sequence beyond end of chromosome ($end > $glen)"
		} else {
			set end $glen
		}
	}
	set pos [expr {$gstart+$start}]
	seek $f $pos
	read $f [expr {$end-$start}]
}

proc loc_compare {loc1 loc2} {
	if {$loc1 eq $loc2} {return 0}
	set temp [list $loc1 $loc2]
	if {[lsort -dict $temp] eq $temp} {return -1} else {return 1}
}

proc chr_compare {chr1 chr2} {
	set chr1 [chr_clip $chr1]
	set chr2 [chr_clip $chr2]
	loc_compare $chr1 $chr2
}

proc lloc_compare {loc1 loc2} {
	loc_compare [join $loc1 " "] [join $loc2 " "]
}
